import hashlib
import os
from configparser import ConfigParser
from sanic.request import Request
from sanic.response import HTTPResponse, redirect

security_cache_path = "./resources/security-cache"
config = ConfigParser()
config.read("./security.ini")


def xss_prevention_middleware(request: Request, response: HTTPResponse):
    """
    Adds a header to all responses that prevents cross site scripting.

    Args:
        response (HTTPResponse): Sanic response parameter.
    """
    response.headers["x-xss-protection"] = "1; mode=block"


def https_redirect_middleware(request: Request):
    """
    Redirects all http requests to https.

    Args:
        request (Request): Sanic request parameter.
    """
    if request.url.startswith("http://"):
        url = request.url.replace("http://", "https://", 1)
        return redirect(url)


def hash_password(password: str):
    """
    Securely hashes passed password to be stored. The function provides PKCS#5 password-based key derivation function 2. It uses HMAC as pseudorandom function.

    Args:
        password (str): Password to be hashed.

    Returns:
        hashed_password
    """
    return hashlib.pbkdf2_hmac(
        "sha512",
        password.encode("utf-8"),
        config["SECURITY"]["SECRET"].encode("utf-8"),
        100000,
    )


def get_ip(request: Request):
    """
    Retrieves ip address from request.

    Args:
        request (Request): Sanic request parameter.

    Returns:
        ip
    """
    return request.remote_addr if request.remote_addr else request.ip


def dir_exists(path: str):
    """
    Checks if path exists and isn't empty, and creates it if it doesn't.

    Args:
         path (str): Path being checked.

    Returns:
        exists
    """
    try:
        os.makedirs(path)
        exists = False
    except FileExistsError:
        exists = os.listdir(path)
    return exists



from sanic import Sanic
from sanic.exceptions import ServerError
from sanic.response import json as sanic_json
from sanic.response import text, file

from sanic_security.core.authentication import (
    register,
    login,
    requires_authentication,
    logout,
)
from sanic_security.core.authorization import require_permissions, require_roles
from sanic_security.core.initializer import initialize_security
from sanic_security.core.models import (
    SecurityError,
    Permission,
    Role,
    CaptchaSession,
    TwoStepSession,
)
from sanic_security.core.recovery import (
    attempt_account_recovery,
    fulfill_account_recovery_attempt,
)
from sanic_security.core.utils import xss_prevention_middleware
from sanic_security.core.verification import (
    requires_captcha,
    request_captcha,
    requires_two_step_verification,
    verify_account,
    request_two_step_verification,
)

app = Sanic("Sanic Security Test Server")


def json(message, data, status_code=200):
    payload = {"message": message, "status_code": status_code, "data": data}
    return sanic_json(payload, status=status_code)


def check_for_empty(form, *args):
    for key, value in form.items():
        if value is not None:
            if not isinstance(value[0], bool) and not value[0] and key not in args:
                raise ServerError(key + " is empty!", 400)


@app.middleware("response")
async def xxs_middleware(request, response):
    """
    Response middleware test.
    """
    xss_prevention_middleware(request, response)


@app.post("api/test/register")
async def on_register(request):
    """
    Registration test without verification or captcha requirements.
    """
    account = await register(request, verified=True)
    return json("Registration Successful!", account.json())


@app.post("api/test/register/verification")
@requires_captcha()
async def on_register_verification(request, captcha_session):
    """
    Registration test with all built-in requirements.
    """
    two_step_session = await register(request)
    await two_step_session.text_code()
    response = json("Registration successful", two_step_session.account.json())
    two_step_session.encode(response, secure=False)
    return response


@app.post("api/test/register/verify")
@requires_two_step_verification()
async def on_verify(request, two_step_session):
    """
    Attempt to verify account and allow access if unverified.
    """
    await verify_account(two_step_session)
    return json("Verification successful!", two_step_session.json())


@app.get("api/test/captcha/img")
async def on_captcha_img(request):
    """
    Retrieves captcha image from captcha session.
    """
    captcha_session = await CaptchaSession().decode(request)
    return await file(captcha_session.get_image())


@app.get("api/test/captcha")
async def on_request_captcha(request):
    """
    Requests captcha session for client.
    """
    captcha_session = await request_captcha(request)
    response = json("Captcha request successful!", captcha_session.json())
    captcha_session.encode(response, secure=False)
    return response


@app.post("api/test/verification/resend")
async def resend_verification_request(request):
    """
    Resends verification code if somehow lost.
    """
    two_step_session = await TwoStepSession().decode(request)
    await two_step_session.text_code()
    return json("Verification code resend successful", two_step_session.json())


@app.post("api/test/verification/request")
@requires_captcha()
async def new_verification_request(request, captcha_session):
    """
    Creates new verification code.
    """
    two_step_session = await request_two_step_verification(request)
    await two_step_session.text_code()
    response = json("Verification request successful", two_step_session.json())
    two_step_session.encode(response, secure=False)
    return response


@app.post("api/test/login")
async def on_login(request):
    """
    User login, creates and encodes authentication session.
    """
    authentication_session = await login(request)
    response = json("Login successful!", authentication_session.account.json())
    authentication_session.encode(response, secure=False)
    return response


@app.post("api/test/logout")
async def on_logout(request):
    """
    User logout, invalidates client authentication session.
    """
    authentication_session = await logout(request)
    response = json("Logout successful!", authentication_session.account.json())
    return response


@app.post("api/test/role/admin")
@requires_authentication()
async def on_create_admin(request, authentication_session):
    """
    Creates 'Admin' and 'Mod' roles to be used for testing role based authorization.
    """
    client = authentication_session.account
    await Role().create(account=client, name="Admin")
    await Role().create(account=client, name="Mod")
    return json("Roles added to your account!", client.json())


@app.post("api/test/perms/admin")
@requires_authentication()
async def on_create_admin_perm(request, authentication_session):
    """
    Creates 'admin:update' and 'admin:add' permissions to be used for testing wildcard based authorization.
    """
    client = authentication_session.account
    await Permission().create(account=client, wildcard="admin:update", decription="")
    await Permission().create(account=client, wildcard="admin:add")
    return json("Permissions added to your account!", client.json())


@app.get("api/test/client")
@requires_authentication()
async def on_test_client(request, authentication_session):
    """
    Retrieves authenticated client username.
    """
    return text("Hello " + authentication_session.account.username + "!")


@app.get("api/test/perm")
@require_permissions("admin:update")
async def on_test_perm(request, authentication_session):
    """
    Tests client wildcard permissions authorization access.
    """
    return text("Admin who can only update gained access!")


@app.get("api/test/role")
@require_roles("Admin", "Mod")
async def on_test_role(request, authentication_session):
    """
    Tests client role authorization access.
    """
    return text("Admin gained access!")


@app.post("api/test/recovery/attempt")
@requires_captcha()
async def on_recovery_attempt(request, captcha_session):
    """
    Attempts to recover account via changing password, requests verification to ensure the recovery attempt was made
    by account owner.
    """
    two_step_session = await attempt_account_recovery(request)
    await two_step_session.text_code()
    response = json(
        "A recovery attempt has been made, please verify account ownership.",
        two_step_session.json(),
    )
    two_step_session.encode(response, secure=False)
    return response


@app.post("api/test/recovery/fulfill")
@requires_two_step_verification()
async def on_recovery_fulfill(request, two_step_session):
    """
    Changes and recovers an account's password once recovery attempt was determined to have been made by account owner.
    """
    await fulfill_account_recovery_attempt(request, two_step_session)
    return json("Account recovered successfully", two_step_session.account.json())


@app.exception(SecurityError)
async def on_error(request, exception):
    return exception.response


if __name__ == "__main__":
    initialize_security(app)
    app.run(host="0.0.0.0", port=8000, debug=True, workers=4)



import functools
from fnmatch import fnmatch

from sanic.request import Request

from sanic_security.core.authentication import authenticate
from sanic_security.core.exceptions import (
    InsufficientPermissionError,
    InsufficientRoleError,
)
from sanic_security.core.models import Role, Permission


async def check_permissions(request: Request, *required_permissions: str):
    """
    Used to determine if the client has sufficient permissions for an action.

    Args:
        request (Request): Sanic request parameter.
        *required_permissions (Tuple[str, ...]):  The permissions required to authorize an action.

    Returns:
        authentication_session

    Raises:
        AccountError
        SessionError
    """
    authentication_session = await authenticate(request)
    client_permissions = await Permission.filter(
        account=authentication_session.account
    ).all()
    for required_permission in required_permissions:
        for client_permission in client_permissions:
            if fnmatch(required_permission, client_permission.wildcard):
                break
        else:
            raise InsufficientPermissionError()
    return authentication_session


async def check_roles(request: Request, *required_roles: str):
    """
    Used to determine if the client has sufficient roles for an action.

    Args:
        request (Request): Sanic request parameter.
        *required_roles (Tuple[str, ...]):  The roles required to authorize an action.

    Returns:
        authentication_session

    Raises:
        AccountError
        SessionError
    """
    authentication_session = await authenticate(request)
    for role in required_roles:
        if await Role.filter(
            account=authentication_session.account, name=role
        ).exists():
            break
    else:
        raise InsufficientRoleError()
    return authentication_session


def require_permissions(*required_permissions: str):
    """
    Used to determine if the client has sufficient permissions for an action.

    Args:
        *required_permissions (Tuple[str, ...]):  The permissions required to authorize an action.

    Example:
        This method is not called directly and instead used as a decorator:

            @app.post('api/authorize')
            @require_permissions('admin:update', 'admin:create')
            async def on_authorize(request, authentication_session):
                return text('User is authorized to update and create data!')

    Raises:
        AccountError
        SessionError
    """

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(request, *args, **kwargs):
            authentication_session = await check_permissions(
                request, *required_permissions
            )
            return await func(request, authentication_session, *args, **kwargs)

        return wrapped

    return wrapper


def require_roles(*required_roles: str):
    """
    Used to determine if the client has sufficient roles for an action.

    Args:
        *required_roles (Tuple[str, ...]): The roles required to authorize an action.

    Example:
        This method is not called directly and instead used as a decorator:

            @app.post('api/authorize')
            @require_roles('Admin', 'Moderator')
            async def on_authorize(request, authentication_session):
                return text('User is authorized with the role Admin or Moderator!')

    Raises:
            AccountError
        SessionError
    """

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(request, *args, **kwargs):
            authentication_session = await check_roles(request, *required_roles)
            return await func(request, authentication_session, *args, **kwargs)

        return wrapped

    return wrapper



from sanic import Sanic

from sanic_security.core.models import CaptchaSession, TwoStepSession
from sanic_security.lib.tortoise import initialize_tortoise


def initialize_security(app: Sanic):
    """
    Starts all initial Sanic Security processes such as initializing tortoise and caches.

    Args:
        app (Sanic): Sanic Framework app.
    """
    initialize_tortoise(app)
    TwoStepSession.initialize_cache(app)
    CaptchaSession.initialize_cache(app)



import functools

from sanic.request import Request

from sanic_security.core.models import (
    Account,
    TwoStepSession,
    CaptchaSession,
    SessionFactory,
    SessionErrorFactory,
)

session_factory = SessionFactory()
session_error_factory = SessionErrorFactory()


async def request_captcha(request: Request):
    """
    Creates a captcha session associated with an account.

    Args:
        request (Request): Sanic request parameter.

    Returns:
        captcha_session
    """
    return await session_factory.get("captcha", request)


async def captcha(request: Request):
    """
    Enforces a captcha to continue execution.

    Args:
        request (Request): Sanic request parameter. All request bodies are sent as form-data with the following arguments: captcha.

    Raises:
        SessionError

    Returns:
        captcha_session
    """
    captcha_session = await CaptchaSession().decode(request)
    session_error_factory.throw(captcha_session)
    await captcha_session.crosscheck_code(request.form.get("captcha"))
    return captcha_session


def requires_captcha():
    """
    Enforces a captcha to continue execution.

    Example:
        This method is not called directly and instead used as a decorator:

            @app.post('api/captcha')
            @requires_captcha()
            async def on_captcha(request, captcha_session):
                return text('User has successfully completed captcha challenge!')

    Raises:
        SessionError
    """

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(request, *args, **kwargs):
            captcha_session = await captcha(request)
            return await func(request, captcha_session, *args, **kwargs)

        return wrapped

    return wrapper


async def request_two_step_verification(request: Request, account: Account = None):
    """
    Creates a two-step session associated with an account.

    Args:
        request (Request): Sanic request parameter.
        account (Account): The account being associated with the verification session. If None, will retrieve account from client Two-Step Session cookie.

    Returns:
         two_step_session
    """
    if account is None:
        two_step_session = await TwoStepSession().decode(request)
        account = two_step_session.account
    return await session_factory.get("twostep", request, account=account)


async def verify_two_step_verification(request: Request):
    """
    Enforces two-step verification to continue action.

    Args:
        request (Request): Sanic request parameter. All request bodies are sent as form-data with the following arguments: code.

    Raises:
        SessionError

    Returns:
         two_step_session
    """
    two_step_session = await TwoStepSession().decode(request)
    session_error_factory.throw(two_step_session)
    await two_step_session.crosscheck_location(request)
    await two_step_session.crosscheck_code(request.form.get("code"))
    return two_step_session


async def verify_account(two_step_session: TwoStepSession):
    """
    Verifies account associated to a two-step session.

    Args:
        two_step_session (TwoStepSession): Two-step session containing account being verified.

    Raises:
        SessionError

    Returns:
         two_step_session
    """
    two_step_session.account.verified = True
    await two_step_session.account.save(update_fields=["verified"])
    return two_step_session


def requires_two_step_verification():
    """
    Enforces two-step verification to continue action.

    Example:
        This method is not called directly and instead used as a decorator:

            @app.post('api/captcha')
            @requires_two_step_verification()
            async def on_two_step_verification(request, two_step_session):
                return text('User has successfully provided the correct verification code from email/sms!')

    Raises:
        SessionError
    """

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(request, *args, **kwargs):
            two_step_session = await verify_two_step_verification(request)
            return await func(request, two_step_session, *args, **kwargs)

        return wrapped

    return wrapper



import setuptools
from setuptools import setup

with open("README.md", "r") as rm:
    long_description = rm.read()

with open("requirements.txt", "r") as rq:
    requirements = [r.strip() for r in rq.readlines()]

setup(
    name='sanic-security',
    version='0.9.13.3',
    packages=setuptools.find_packages(),
    url='https://github.com/sunset-developer/sanic-security',
    license='GNU General Public License v3.0',
    author='sunset-developer',
    author_email='aidanstewart@sunsetdeveloper.com',
    description='A powerful, simple, and async security library for Sanic.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements
)



from sanic.exceptions import SanicException
from sanic.response import json


class SecurityError(SanicException):
    """
    Sanic Security related error.

    Attributes:
        response (HTTPResponse): Security Error json response.

    Args:
        message (str): Human readable error message.
        code (int): HTTP Error code.
    """

    def __init__(self, message: str, code: int):
        self.response = json(
            {
                "message": "An error has occurred!",
                "error_code": code,
                "data": {"error": self.__class__.__name__, "summary": message},
            },
            status=code,
        )
        super().__init__(message, code)


class NotFoundError(SecurityError):
    """
    Raised when a model can't be found in the database.
    Args:
        message (str): Human readable error message.
    """

    def __init__(self, message):
        super().__init__(message, 404)


class DeletedError(SecurityError):
    """
    Raised when a model in the database has been marked deleted.
    Args:
        message (str): Human readable error message.
    """

    def __init__(self, message):
        super().__init__(message, 404)


class AccountError(SecurityError):
    """
    An account related error.

    Args:
        message (str): Human readable error message.
        code (int): HTTP Error code.
    """

    def __init__(self, message, code):
        super().__init__(message, code)


class ExistsError(AccountError):
    def __init__(self):
        super().__init__("Account with this email or phone number already exists.", 409)


class TooManyCharsError(AccountError):
    def __init__(self):
        super().__init__("Email, username, or phone number is too long.", 400)


class InvalidEmailError(AccountError):
    def __init__(self):
        super().__init__("Please use a valid email format such as you@mail.com.", 400)


class DisabledError(AccountError):
    def __init__(self):
        super().__init__("This account has been disabled.", 401)


class PasswordMismatchError(AccountError):
    def __init__(self):
        super().__init__("The password provided does not match account password.", 401)


class UnverifiedError(AccountError):
    def __init__(self):
        super().__init__("Account requires verification.", 401)


class SessionError(SecurityError):
    """
    A session related error.

    Args:
        message (str): Human readable error message.
        code (int): HTTP Error code.
    """

    def __init__(self, message, code):
        super().__init__(message, code)


class DecodingError(SessionError):
    def __init__(self, exception):
        super().__init__("Session cookie could not be decoded. " + str(exception), 400)


class InvalidError(SessionError):
    def __init__(self):
        super().__init__("Session is invalid.", 401)


class ExpiredError(SessionError):
    def __init__(self):
        super().__init__("Session has expired", 401)


class CrosscheckError(SessionError):
    def __init__(self):
        super().__init__("Session crosschecking attempt was incorrect", 401)


class MaximumAttemptsError(SessionError):
    def __init__(self):
        super().__init__(
            "You've reached the maximum amount of attempts for this session.", 401
        )


class UnknownLocationError(SessionError):
    def __init__(self):
        super().__init__("Session in an unknown location.", 401)


class InsufficientRoleError(SecurityError):
    def __init__(self):
        super().__init__("Insufficient roles required for this action.", 403)


class InsufficientPermissionError(SecurityError):
    def __init__(self):
        super().__init__("Insufficient permissions required for this action.", 403)
